from prelude import p as Prelude

FalseValue = '#f'
TrueValue = '#t'
NilValue = 'nil'
QuoteSymbol = 'quote'
IfSymbol = 'if'
BeginSymbol = 'begin'
SetSymbol = 'set!'
DefineSymbol = 'define'
LambdaSymbol = 'lambda'
MuSymbol = 'mu'


def Cons(X, Y):
    return [X, Y]

def Car(X):
    return X[0]

def Cdr(X):
    return X[1]

def Add(X, Y):
    return X + Y

def Sym(X):
    return str(X)

def Eq(X, Y):
    if X is Y:
        return TrueValue
    else:
        return FalseValue

def Pair(X):
    if isinstance(X, list):
        return TrueValue
    else:
        return FalseValue
    

def EvaluateIfExpression(Predicate, Consequent, Alternative, Environment, Continuation):
    def IfContinuation(Value, _, __):
        if Value == FalseValue:
            return Alternative, Environment, Continuation
        else:
            return Consequent, Environment, Continuation
    return Predicate, Environment, IfContinuation

def EvaluateBeginExpression(Body, Environment, Continuation):
    def BeginContinuation(_, __, ___):
        return EvaluateBeginExpression(Body[1], Environment, Continuation)
    if Cdr(Body) == NilValue:
        return Car(Body), Environment, Continuation
    else:
        return Car(Body), Environment, BeginContinuation

def EvaluateSetExpression(Name, Expression, Environment, Continuation):
    def SetContinuation(Value, _, __):
        Environment.Set(Name, Value)
        return NilValue, Environment, Continuation
    return Expression, Environment, SetContinuation

def EvaluateDefineExpression(Name, Expression, Environment, Continuation):
    def DefineContinuation(Value, _, __):
        Environment.Define(Name, Value)
        return NilValue, Environment, Continuation
    return Expression, Environment, DefineContinuation

def EvaluateApplication(FunctionExpression, ArgumentExpressions, Environment, Continuation):
    def ApplicationContinuation(Value, _, __):
        return Value(ArgumentExpressions, Environment, Continuation)
    return FunctionExpression, Environment, ApplicationContinuation

def EvaluateLambdaAbstraction(Parameters, Body, EnclosingEnvironment, Continuation):
    def Function(ArgumentExpressions, CallingEnvironment, CallContinuation):
        def ParameterContinuation(ArgumentExpressions, ArgumentValues):
            if ArgumentExpressions == NilValue:
                EnvironmentFrame = {}
                ParameterTraverse = Parameters
                while isinstance(ParameterTraverse, list):
                    EnvironmentFrame[Car(ParameterTraverse)] = Car(ArgumentValues)
                    ArgumentValues = ArgumentValues[1:]
                    ParameterTraverse = Cdr(ParameterTraverse)
                if ParameterTraverse != NilValue:
                    CollectedArguments = NilValue
                    for Argument in reversed(ArgumentValues):
                        CollectedArguments = Cons(Argument, CollectedArguments)
                    EnvironmentFrame[ParameterTraverse] = CollectedArguments
                return EvaluateBeginExpression(Body,
                                               EnclosingEnvironment.Extend(EnvironmentFrame),
                                               CallContinuation)
            else:
                def ArgumentContinuation(Value, _, __):
                    return ParameterContinuation(Cdr(ArgumentExpressions),
                                                 [*ArgumentValues, Value])
                return Car(ArgumentExpressions), CallingEnvironment, ArgumentContinuation
        return ParameterContinuation(ArgumentExpressions, [])
    return Function, EnclosingEnvironment, Continuation

def EvaluateMuAbstraction(Parameters, Body, EnclosingEnvironment, Continuation):
    def Macro(ParameterExpressions, CallingEnvironment, CallContinuation):
        ParameterTraverse = Parameters
        EnvironmentFrame = {}
        while isinstance(ParameterTraverse, list):
            EnvironmentFrame[Car(ParameterTraverse)] = Car(ParameterExpressions)
            ParameterTraverse = Cdr(ParameterTraverse)
            ParameterExpressions = Cdr(ParameterExpressions)
        if ParameterTraverse != NilValue:
            EnvironmentFrame[ParameterTraverse] = ParameterExpressions
        def EvalContinuation(Value, _, __):
            return Value, CallingEnvironment, CallContinuation
        return EvaluateBeginExpression(Body,
                                       EnclosingEnvironment.Extend(EnvironmentFrame),
                                       EvalContinuation)
    return Macro, EnclosingEnvironment, Continuation

def MakePrimitive(NativeFunction):
    def Function(ParameterExpressions, CallingEnvironment, CallContinuation):
        def PrimitiveParameterContinuation(ParameterExpressions, ArgumentValues):
            if ParameterExpressions == NilValue:
                return CallContinuation(NativeFunction(*ArgumentValues),
                                        CallingEnvironment,
                                        CallContinuation)
            else:
                def PrimitiveArgumentContinuation(Value, _, __):
                    return PrimitiveParameterContinuation(Cdr(ParameterExpressions),
                                                          [*ArgumentValues, Value])
                return Car(ParameterExpressions), CallingEnvironment, PrimitiveArgumentContinuation
        return PrimitiveParameterContinuation(ParameterExpressions, [])
    return Function


class Environment:
    def __init__(Self, Parent=None, Bindings=None):
        if Bindings == None:
            Bindings = {}
        Self.Parent = Parent
        Self.Bindings = Bindings

    def Get(Self, Name):
        while Self != None:
            if Name in Self.Bindings:
                return Self.Bindings[Name]
            Self = Self.Parent

    def Set(Self, Name, Value):
        while Self != None:
            if Name in Self.Bindings:
                Self.Bindings[Name] = Value
                return
            Self = Self.Parent

    def Define(Self, Name, Value):
        Self.Bindings[Name] = Value

    def Extend(Self, Bindings):
        return Environment(Self, Bindings)


def ReturnContinuation(Expression, Environment, Continuation):
    return Expression, None, None


def Evaluate(Expression, Environment, Continuation):
    while Continuation != None:
        if isinstance(Expression, str):
            Expression, Environment, Continuation = Continuation(Environment.Get(Expression),
                                                                 Environment, Continuation)
        elif not isinstance(Expression, list):
            Expression, Environment, Continuation = Continuation(Expression,
                                                                 Environment,
                                                                 Continuation)
        elif Car(Expression) == QuoteSymbol:
            Quotation = Car(Cdr(Expression))
            Expression, Environment, Continuation = Continuation(Quotation,
                                                                 Environment,
                                                                 Continuation)
        elif Car(Expression) == IfSymbol:
            Predicate = Car(Cdr(Expression))
            Consequent = Car(Cdr(Cdr(Expression)))
            Alternative = Car(Cdr(Cdr(Cdr(Expression))))
            Expression, Environment, Continuation = EvaluateIfExpression(Predicate,
                                                                         Consequent,
                                                                         Alternative,
                                                                         Environment,
                                                                         Continuation)
        elif Car(Expression) == BeginSymbol:
            Body = Cdr(Expression)
            Expression, Environment, Continuation = EvaluateBeginExpression(Body,
                                                                            Environment,
                                                                            Continuation)
        elif Car(Expression) == SetSymbol:
            Name = Car(Cdr(Expression))
            Value = Car(Cdr(Cdr(Expression)))
            Expression, Environment, Continuation = EvaluateSetExpression(Name,
                                                                          Value,
                                                                          Environment,
                                                                          Continuation)
        elif Car(Expression) == DefineSymbol:
            Name = Car(Cdr(Expression))
            Value = Car(Cdr(Cdr(Expression)))
            Expression, Environment, Continuation = EvaluateDefineExpression(Name,
                                                                             Value,
                                                                             Environment,
                                                                             Continuation)
        elif Car(Expression) == LambdaSymbol:
            Parameters = Car(Cdr(Expression))
            Body = Cdr(Cdr(Expression))
            Expression, Environment, Continuation = EvaluateLambdaAbstraction(Parameters,
                                                                              Body,
                                                                              Environment,
                                                                              Continuation)
        elif Car(Expression) == MuSymbol:
            Parameters = Car(Cdr(Expression))
            Body = Cdr(Cdr(Expression))
            Expression, Environment, Continuation = EvaluateMuAbstraction(Parameters,
                                                                          Body,
                                                                          Environment,
                                                                          Continuation)
        else:
            Applicant = Car(Expression)
            Arguments = Cdr(Expression)
            Expression, Environment, Continuation = EvaluateApplication(Applicant,
                                                                        Arguments,
                                                                        Environment,
                                                                        Continuation)
    return Expression


def Consify(List):
    if type(List) != list:
        return List
    if List[-2:-1] == [...]:
        ConsList = List[-1]
        StartIndex = -3
    else:
        ConsList = NilValue
        StartIndex = -1
    for Element in List[StartIndex::-1]:
        ConsList = Cons(Consify(Element), ConsList)
    return ConsList

def Flatten(List, Recursing=False):
    if List == NilValue:
        if Recursing:
            return []
        else:
            return NilValue
    elif type(List) != list:
        if Recursing:
            return [..., List]
        else:
            return List
    else:
        return [Flatten(Car(List))] + Flatten(Cdr(List), Recursing=True)


def AnyLisp(Expression):
    return Flatten(Evaluate(Consify(Expression), GlobalEnvironment, ReturnContinuation))

GlobalEnvironment = Environment().Extend({ 
    'cons':         MakePrimitive(Cons),
    'car':          MakePrimitive(Car),
    'cdr':          MakePrimitive(Cdr),
    'eq?':          MakePrimitive(Eq),
    '+':            MakePrimitive(Add),
    'pair?':        MakePrimitive(Pair),      
    'sym':          MakePrimitive(Sym),
})

AnyLisp(Prelude)

